namespace dnnl {
namespace impl {
namespace gpu {
namespace ocl {
const char *reduction_utils_header = R"==(/*******************************************************************************)==""\n"
R"==(* Copyright 2023 Intel Corporation)==""\n"
R"==(*)==""\n"
R"==(* Licensed under the Apache License, Version 2.0 (the "License");)==""\n"
R"==(* you may not use this file except in compliance with the License.)==""\n"
R"==(* You may obtain a copy of the License at)==""\n"
R"==(*)==""\n"
R"==(* http:)==""\n"
R"==(*)==""\n"
R"==(* Unless required by applicable law or agreed to in writing, software)==""\n"
R"==(* distributed under the License is distributed on an "AS IS" BASIS,)==""\n"
R"==(* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.)==""\n"
R"==(* See the License for the specific language governing permissions and)==""\n"
R"==(* limitations under the License.)==""\n"
R"==(*******************************************************************************/)==""\n"
R"==(#ifndef GPU_REDUCTION_UTILS_H)==""\n"
R"==(#define GPU_REDUCTION_UTILS_H)==""\n"
R"==(#include <assert.h>)==""\n"
R"==(#include <sstream>)==""\n"
R"==(#include <vector>)==""\n"
R"==(#include "common/c_types_map.hpp")==""\n"
R"==(#include "gpu/block_structure.hpp")==""\n"
R"==(namespace dnnl {)==""\n"
R"==(namespace impl {)==""\n"
R"==(namespace gpu {)==""\n"
R"==(namespace ocl {)==""\n"
R"==(struct zero_padding_t {)==""\n"
R"==(zero_padding_t(const int dim_idx, const dim_t data_size,)==""\n"
R"==(const dim_t outer_stride, const dim_t outer_size,)==""\n"
R"==(const dim_t inner_stride, const dim_t inner_size))==""\n"
R"==(: dim_idx(dim_idx))==""\n"
R"==(, data_size(data_size))==""\n"
R"==(, outer_stride(outer_stride))==""\n"
R"==(, outer_size(outer_size))==""\n"
R"==(, inner_stride(inner_stride))==""\n"
R"==(, inner_size(inner_size) {})==""\n"
R"==(zero_padding_t()==""\n"
R"==(const dim_t data_size, const block_t &outer, const block_t &inner))==""\n"
R"==(: data_size(data_size))==""\n"
R"==(, outer_stride(outer.stride))==""\n"
R"==(, outer_size(outer.block))==""\n"
R"==(, inner_stride(inner.stride))==""\n"
R"==(, inner_size(inner.block) {)==""\n"
R"==(assert(outer.dim_idx == inner.dim_idx);)==""\n"
R"==(assert(outer.block * inner.block >= data_size);)==""\n"
R"==(dim_idx = outer.dim_idx;)==""\n"
R"==(})==""\n"
R"==(std::string str() const {)==""\n"
R"==(std::stringstream os;)==""\n"
R"==(os << dim_idx << ": ";)==""\n"
R"==(os << "(idx / " << inner_stride << ") % " << inner_size;)==""\n"
R"==(os << " + " << inner_size << " * (";)==""\n"
R"==(os << "(idx / " << outer_stride << ") % " << outer_size;)==""\n"
R"==(os << ") < " << data_size;)==""\n"
R"==(return os.str();)==""\n"
R"==(})==""\n"
R"==(dim_t dim_idx;)==""\n"
R"==(dim_t data_size;)==""\n"
R"==(dim_t outer_stride, outer_size;)==""\n"
R"==(dim_t inner_stride, inner_size;)==""\n"
R"==(};)==""\n"
R"==(class reduction_subproblem_t {)==""\n"
R"==(public:)==""\n"
R"==(reduction_subproblem_t()==""\n"
R"==(dim_t inner_size, dim_t reduction_size, dim_t outer_size))==""\n"
R"==(: inner_block(0, inner_size, 1))==""\n"
R"==(, reduction_block(1, reduction_size, inner_size))==""\n"
R"==(, outer_block(2, outer_size, inner_size * reduction_size) {})==""\n"
R"==(reduction_subproblem_t(std::vector<block_t> blocks, size_t red_start_idx,)==""\n"
R"==(size_t red_end_idx, int reduced_mask))==""\n"
R"==(: inner_block(0, 1, 1))==""\n"
R"==(, reduction_block(1, 1, blocks[red_start_idx].stride))==""\n"
R"==(, outer_block(2, 1,)==""\n"
R"==(red_end_idx == blocks.size())==""\n"
R"==(? blocks.back().stride * blocks.back().block)==""\n"
R"==(: blocks[red_end_idx].stride) {)==""\n"
R"==(dim_t ignored_inner_elems = 1;)==""\n"
R"==(for (size_t i = 0; i < blocks.size(); i++) {)==""\n"
R"==(dim_t block_size = blocks[i].block;)==""\n"
R"==(if (i < red_start_idx) {)==""\n"
R"==(if (reduced_mask & (1 << blocks[i].dim_idx)) {)==""\n"
R"==(ignored_inner_elems *= block_size;)==""\n"
R"==(} else {)==""\n"
R"==(inner_block.block *= block_size;)==""\n"
R"==(})==""\n"
R"==(} else if (i < red_end_idx) {)==""\n"
R"==(assert(reduced_mask & (1 << blocks[i].dim_idx));)==""\n"
R"==(reduction_block.block *= block_size;)==""\n"
R"==(} else {)==""\n"
R"==(outer_block.block *= block_size;)==""\n"
R"==(})==""\n"
R"==(})==""\n"
R"==(reduction_block.stride /= ignored_inner_elems;)==""\n"
R"==(outer_block.stride /= ignored_inner_elems;)==""\n"
R"==(})==""\n"
R"==(std::string str() const {)==""\n"
R"==(std::stringstream os;)==""\n"
R"==(os << "subproblem:" << std::endl;)==""\n"
R"==(os << "outer: " << outer_block.str() << std::endl;)==""\n"
R"==(os << "reduction: " << reduction_block.str() << std::endl;)==""\n"
R"==(os << "inner: " << inner_block.str() << std::endl;)==""\n"
R"==(os << " -- src zero pads: " << src_zpads.size() << std::endl;)==""\n"
R"==(for (const auto &zp : src_zpads) {)==""\n"
R"==(os << " + " << zp.str() << std::endl;)==""\n"
R"==(})==""\n"
R"==(os << " -- dst zero pads: " << dst_zpads.size() << std::endl;)==""\n"
R"==(for (const auto &zp : dst_zpads) {)==""\n"
R"==(os << " + " << zp.str() << std::endl;)==""\n"
R"==(})==""\n"
R"==(return os.str();)==""\n"
R"==(})==""\n"
R"==(block_t inner_block;)==""\n"
R"==(block_t reduction_block;)==""\n"
R"==(block_t outer_block;)==""\n"
R"==(std::vector<zero_padding_t> src_zpads;)==""\n"
R"==(std::vector<zero_padding_t> dst_zpads;)==""\n"
R"==(};)==""\n"
R"==(status_t generate_reduction_phases(const memory_desc_t *src,)==""\n"
R"==(const memory_desc_t *dst, std::vector<reduction_subproblem_t> &subprbs);)==""\n"
R"==(} )==""\n"
R"==(} )==""\n"
R"==(} )==""\n"
R"==(} )==""\n"
R"==(#endif)==""\n"
R"==()==";
}
}
}
}